"""
    这里是丢弃错误，用来抛掉不要的数据
"""


# 异常：丢弃请求
class DropRequestException(Exception):
    def __init__(self, *args):
        super(DropRequestException, self).__init__(*args)


# 异常：丢弃 item
class DropItemException(Exception):
    def __init__(self, *args):
        super(DropItemException, self).__init__(*args)



"""
    邮件发送
"""
import zmail
from palp import settings


def send_email(sender, subject, content):
    """

    :param sender: 发送者
    :param subject: 主题
    :param content: 发送内容
    :return:
    """
    server = zmail.server(settings.EMAIL_USER, settings.EMAIL_PWD)

    info = {
        'from': sender,
        'subject': subject,
        'content_text': content,
    }

    server.send_mail(settings.EMAIL_RECEIVER, info)



"""
    基础的 item
"""
import json
from loguru import logger


class BaseItem:
    __table_name__ = None  # 表名

    def to_dict(self) -> dict:
        """
        转化为 dict

        :return:
        """
        item = {}

        for key, value in self.__dict__.items():
            item[key] = value

        return item

    def to_json(self, **kwargs) -> str:
        """
        转化为 json

        :param kwargs: json 参数
        :return:
        """
        kwargs.setdefault('ensure_ascii', False)

        return json.dumps(self.to_dict(), **kwargs)

    def __setattr__(self, key, value):
        if not isinstance(key, str):
            logger.warning("item key 必须是一个字符串")
            key = str(key)

        self.__dict__[key] = value

    def __str__(self):
        return f"<{self.__class__.__name__} table:{self.__class__.__table_name__} item:{self.to_dict()}>"



"""
    item 处理

    可自定义缓存数量，一次性进入 pipeline_save 函数入库
"""
import threading
from abc import abstractmethod
from palp.item.item_base import BaseItem


class BaseItemBuffer(threading.Thread):
    PIPELINE = []

    @classmethod
    def from_settings(cls):
        """
        加载设置

        :return:
        """
        pass

    @abstractmethod
    def run(self) -> None:
        """
        入口函数

        :return:
        """
        pass

    @abstractmethod
    def pipeline_in(self, item: BaseItem):
        """
        存放、清洗 item

        :param item:
        :return:
        """
        pass

    @abstractmethod
    def pipeline_save(self):
        """
        处理入库

        :return:
        """
        pass

    @abstractmethod
    def pipeline_close(self):
        """
        spider 结束时运行

        :return:
        """
        pass

    @property
    @abstractmethod
    def buffer_size(self):
        """
        返回 buffer 大小

        :return:
        """
        return



"""
    三种内存队列
"""
import queue
from palp.network.request import Request
from palp.sequence.sequence_base import BaseSequence


# 先进先出队列
class FIFOSequence(BaseSequence):
    def __init__(self):
        self.queue = queue.Queue()

    def put(self, obj, timeout: int = None):
        """
        添加任务

        :param timeout:
        :param obj: 数据
        :return:
        """
        self.queue.put(obj, timeout=timeout)

    def get(self, timeout: int = None):
        """
        获取任务
        :return:
        """
        try:
            item = self.queue.get(timeout=timeout)
            return item
        except queue.Empty:
            return

    def empty(self):
        """
        判断队列是否为空

        :return:
        """
        return self.queue.empty()


# 后进先出队列
class LIFOSequence(FIFOSequence):
    def __init__(self):
        super().__init__()
        self.queue = queue.LifoQueue()


# 优先级队列
class PrioritySequence(LIFOSequence):
    def __init__(self):
        super().__init__()
        self.queue = queue.PriorityQueue()

    def put(self, obj, timeout: int = None):
        """
        添加任务

        :param timeout:
        :param obj: 数据
        :return:
        """
        if isinstance(obj, Request):
            level = obj.level
        else:
            level = 10

        self.queue.put([level, obj], timeout=timeout)

    def get(self, timeout: int = None):
        """
        获取任务（这里的 put 都必须是 [level，value] 的形式）

        :return:
        """
        try:
            item = self.queue.get(timeout=timeout)
            return item[-1]
        except queue.Empty:
            return



"""
    spider 基类
"""
import sys
import types
import inspect
import requests
import traceback
from pathlib import Path
from loguru import logger
from palp import settings
from threading import Thread
from palp.parser import Parser
from typing import List, Union
from palp.network.request import Request
from palp.network.response import Response
from requests.cookies import RequestsCookieJar
from palp.tool.short_module import import_module
from palp.exception.exception_error import NotGeneratorFunctionError
from palp.sequence.sequence_memory import FIFOSequence as FIFOSequenceMemory
from palp.sequence.sequence_redis_item import FIFOSequence as FIFOSequenceRedis


class BaseSpider(Thread):
    SETTINGS: Union[types.ModuleType, dict] = None  # 用来指定爬虫单独的配置，可以是一个引入的 settings.py 模块 或 字典
    SPIDER_MIDDLEWARE: List[types.ModuleType] = []

    def __init__(self, thread_count: int = None, request_filter: bool = False, item_filter: bool = False):
        super().__init__()
        # 加载设置
        self.from_settings()

        # 加载快捷的设置
        if thread_count:
            setattr(settings, 'SPIDER_THREAD_COUNT', thread_count)
        if request_filter:
            setattr(settings, 'REQUEST_FILTER', request_filter)
        if item_filter:
            setattr(settings, 'ITEM_FILTER', item_filter)

        # 获取队列
        queue_module = settings.REQUEST_QUEUE[settings.SPIDER_TYPE][settings.REQUEST_MODE]
        self._queue = import_module(queue_module)[0]  # 存储 请求
        if settings.SPIDER_TYPE == 1:
            self._queue_item = FIFOSequenceMemory()  # 存储 item
        else:
            self._queue_item = FIFOSequenceRedis()

        self._parser_list = []  # 存储所有的解析器
        self._item_buffer = None  # item buffer 线程
        self.spider_done = False  # 爬虫是否运行结束
        self.spider_master = False  # 是否是主机器

    @classmethod
    def from_settings(cls) -> None:
        """
        spider 启动时的自定义配置解析自定义的配置文件

        :return:
        """
        # 修改日志
        if settings.LOG_SAVE:
            log_path = Path(settings.LOG_PATH).joinpath(cls.__name__)
            log_path.mkdir(parents=True, exist_ok=True)

            logger.remove()
            logger.add(
                sink=log_path.joinpath(f'{cls.__name__}.log'),
                enqueue=True,
                backtrace=True,
                diagnose=True,
                rotation=settings.LOG_ROTATION,
                compression=settings.LOG_COMPRESSION,
                level=settings.LOG_LEVEL,
                retention=settings.LOG_RETENTION,
            )

            # 让日志同时能够输出
            if settings.LOG_SHOW:
                logger.add(sys.stdout, level=settings.LOG_LEVEL)

        # 处理每个爬虫单独的设置
        if isinstance(cls.SETTINGS, types.ModuleType):
            for key, value in cls.SETTINGS.__dict__.items():
                if not key.startswith('__'):
                    setattr(settings, key, value)
        elif isinstance(cls.SETTINGS, dict):
            for key, value in cls.SETTINGS.items():
                setattr(settings, key, value)
        elif cls.SETTINGS is None:
            pass
        else:
            logger.warning("不受支持的设置格式，仅支持模块或者字典！")

        # 引入中间件
        cls.SPIDER_MIDDLEWARE = import_module(settings.SPIDER_MIDDLEWARE)

        # 判断是否使用了连接，预先导入
        if settings.SPIDER_TYPE == 2:
            from palp.conn.conn_redis import Redis

    def start_requests(self) -> None:
        """
        初始请求的发出

        @result: yield Request()
        """
        pass

    def parse(self, request: Request, response: Response) -> None:
        """
        默认的解析函数（仅 start_requests 默认的）

        @result:
        """
        pass

    def distribute_task(self) -> None:
        """
        刚开始的分发任务

        :return:
        """
        # 校验是否是生成器函数
        if not inspect.isgeneratorfunction(self.start_requests):
            raise NotGeneratorFunctionError("start_requests 必须 yield Request!")

        # 检查起始函数发出请求
        for request in self.start_requests():
            if not isinstance(request, Request):
                raise ValueError("start_requests 仅支持 yield Request")

            # 起始函数无 callback 默认添加
            if request.callback is None:
                request.callback = self.parse

            # 先改成名字，不然后续无法序列化
            request.callback = request.callback.__name__

            # 为每一个起始函数添加一个 session、cookie_jar
            request.session = requests.session()
            request.cookie_jar = RequestsCookieJar()

            self._queue.put(request)

    def wait_parser_done(self) -> None:
        """
        等待线程执行结束

        :return:
        """
        while True:
            if self.all_parser_is_waiting():
                logger.debug("所有线程都已挂起，即将停止")
                self.stop_all_parser()
                break
            elif self.all_parser_is_done():
                break

        self.spider_done = True

    def wait_item_buffer_done(self) -> None:
        """
        等待 item buffer 结束

        :return:
        """
        while True:
            if not self._item_buffer.is_alive():
                break

    def stop_all_parser(self):
        """
        停止所有 parser 运行

        :return:
        """
        for p in self._parser_list:
            p.stop = True

        while True:
            if self.all_parser_is_done():
                break

    def all_parser_is_waiting(self) -> bool:
        """
        是否所有 parser 都没事干

        :return:
        """
        waiting_status = True

        for p in self._parser_list:
            if not p.waiting:
                waiting_status = False
                break

        return waiting_status

    def all_parser_is_done(self) -> bool:
        """
        判断所有 parser 是否已经结束

        :return:
        """

        done_status = True

        for p in self._parser_list:
            if p.is_alive():
                done_status = False
                break

        return done_status

    def spider_logic(self) -> None:
        """
        子类实现爬虫逻辑的地方

        :return:
        """
        pass

    def run(self) -> None:
        """
        启动入口

        :return:
        """
        try:
            # spider 开始执行前的处理
            for middleware in self.__class__.SPIDER_MIDDLEWARE:
                middleware.spider_start(self)

            # 启动对应的线程
            Parser.from_settings()
            for i in range(settings.SPIDER_THREAD_COUNT):
                parser = Parser(q=self._queue, q_item=self._queue_item, spider=self)
                self._parser_list.append(parser)
                parser.start()

            # spider 逻辑
            self.spider_logic()

        except Exception as e:
            for middleware in self.__class__.SPIDER_MIDDLEWARE:
                middleware.spider_error(self, e.__class__.__name__, traceback.format_exc())
        finally:
            # 等待线程执行完毕
            self.wait_parser_done()

            # 等待 item buffer 执行完毕
            if settings.SPIDER_TYPE == 1 or (settings.SPIDER_TYPE == 2 and self.spider_master):
                self.wait_item_buffer_done()

            logger.debug("无任务，爬虫结束")

            # spider 结束之后的处理
            for middleware in self.__class__.SPIDER_MIDDLEWARE:
                middleware.spider_close(self)

    @property
    def name(self) -> str:
        """
        获取 spider 的名字

        :return:
        """

        return self.__class__.__name__



"""
    redis 连接
    redis 锁

    RedisLock：所有使用该锁的都会等待锁，直到获取并执行
    RedisLockNoWait：所有使用该锁的，获取到锁才会执行，获取不到就不执行

    注意：默认锁住的最大时间是 60s，超过则释放

    使用示例：
        with RedisLock(lock_name=""):
            ...

        with RedisLockNoWait(lock_name=") as lock:
            if lock.lock_success:
                ...
"""
import redis
import uuid
import rediscluster
from typing import Union
from palp import settings


# redis 连接
class Redis:
    CONNECTION: Union[rediscluster.RedisCluster, redis.Redis] = None

    @classmethod
    def create_connection(cls):
        """
        创建 redis 连接

        :return:
        """
        if settings.REDIS_CLUSTER_NODES:
            pool = rediscluster.ClusterBlockingConnectionPool(
                startup_nodes=settings.REDIS_CLUSTER_NODES,
                password=settings.REDIS_PWD,
                max_connections=settings.REDIS_MAX_CONNECTIONS,
                **settings.REDIS_POOL_CONFIG
            )

            cls.CONNECTION = rediscluster.RedisCluster(connection_pool=pool, **settings.REDIS_CONFIG)
        else:
            pool = redis.BlockingConnectionPool(
                host=settings.REDIS_HOST,
                port=settings.REDIS_PORT,
                password=settings.REDIS_PWD,
                db=settings.REDIS_DB,
                **settings.REDIS_POOL_CONFIG
            )

            cls.CONNECTION = redis.Redis(connection_pool=pool, decode_responses=True, **settings.REDIS_CONFIG)

    @classmethod
    def conn(cls) -> Union[rediscluster.RedisCluster, redis.Redis]:
        """
        用来获取连接

        :return:
        """
        if cls.CONNECTION is None:
            cls.create_connection()

        return cls.CONNECTION


# redis 等待锁
class RedisLock:
    def __init__(
            self,
            lock_name: str,
            block_timeout: int = None,
            wait_timeout: int = None,
            frequency: float = None
    ):
        """

        :param lock_name: 锁的名字
        :param block_timeout: 锁住的最大时间
        :param wait_timeout: 等待锁的最大时间
        :param frequency: 检查锁的频率
        """
        self._lock_name = lock_name
        self._block_timeout = block_timeout or 60
        self._wait_timeout = wait_timeout
        self._frequency = frequency or 0.1

        self._lock = Redis.conn().lock(
            name=self._lock_name,
            timeout=self._block_timeout,
            blocking_timeout=self._wait_timeout,
            sleep=self._frequency
        )

    def acquire(self):
        """
        获取锁

        :return:
        """
        return self._lock.acquire()

    def release(self):
        """
        释放锁

        :return:
        """
        return self._lock.release()

    def __enter__(self):
        self.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()


# redis 不等待锁
class RedisLockNoWait:
    def __init__(self, lock_name: str, block_timeout: int = None, auto_lock: bool = True):
        """

        :param lock_name: 锁的名字
        :param block_timeout: 锁住的最大时间
        :param auto_lock: 是否 with 时自动上锁
        """
        self._lock_name = lock_name
        self._block_timeout = block_timeout or 60
        self._auto_lock = auto_lock

        self.lock_success = False  # 是否成功上锁
        self._token = uuid.uuid1().hex.encode()

    def acquire(self) -> bool:
        """
        获取锁

        :return:
        """

        self.lock_success = bool(Redis.conn().set(
            name=self._lock_name,
            value=self._token,
            nx=True,
            ex=self._block_timeout
        ))

        return self.lock_success

    def release(self) -> bool:
        """
        释放锁

        :return:
        """
        if Redis.conn().get(self._lock_name) == self._token:
            return bool(Redis.conn().delete(self._lock_name))

        return False

    def exists(self) -> bool:
        """
        是否含有锁

        :return:
        """
        return bool(Redis.conn().exists(self._lock_name))

    def __enter__(self):
        self._auto_lock and self.acquire()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.lock_success and self.release()



"""
    处理爬虫 yield 的请求、item 处理
"""
import inspect
import traceback
from urllib.parse import urlparse

from loguru import logger
from palp import settings
from threading import Thread
from palp.network.request import Request
from palp.item.item_base import BaseItem
from palp.exception import exception_drop
from palp.tool.short_module import import_module
from palp.sequence.sequence_base import BaseSequence


class Parser(Thread):
    REQUEST_MIDDLEWARE = []

    def __init__(self, q: BaseSequence, q_item: BaseSequence, spider):
        """

        :param q: request 队列
        :param q_item: item 队列
        :param spider: 运行的爬虫
        """
        super().__init__()
        self.queue = q
        self.queue_item = q_item
        self.spider = spider

        self.waiting = False  # parser 执行状态，如果拿到空任务就是在等待
        self.stop = False  # 是否暂停

    @classmethod
    def from_settings(cls):
        """
        导入过滤设置

        :return:
        """
        request_filter_middleware = import_module(
            settings.REQUEST_FILTER_MIDDLEWARE[settings.SPIDER_TYPE][settings.FILTERING_MODE])
        cls.REQUEST_MIDDLEWARE = request_filter_middleware + import_module(settings.REQUEST_MIDDLEWARE)

    def run(self):
        """
        函数执行入口

        :return:
        """
        while not self.stop:
            task = self.queue.get(timeout=1)

            try:
                if task is None:
                    self.waiting = True
                    continue

                # 将 requests 还原
                elif isinstance(task, Request) and task.callback and hasattr(self.spider, str(task.callback)):
                    task.callback = eval(f"self.spider.{task.callback}")

                self.waiting = False
                self.parse_task(task=task)
            except exception_drop.DropRequestException:
                logger.debug(f"丢弃请求：{task}")
            except Exception as e:
                # spider 报错处理
                for middleware in self.spider.SPIDER_MIDDLEWARE:
                    middleware.spider_error(self.spider, e.__class__.__name__, traceback.format_exc())

    def parse_task(self, task, request: Request = None):
        """
        解析处理任务

        :param task: 任务
        :param request: 上一个的请求
        :return:
        """
        # 判断 yield 的类型
        if isinstance(task, Request):
            if request:
                self.add_new_request(task, request)
            else:
                self.run_requests(task)
        elif isinstance(task, BaseItem):
            self.queue_item.put(task)
        else:
            logger.warning(f"捕获到非法 yield：{task}")

    def add_new_request(self, new_request: Request, old_request: Request):
        """
        添加新的请求到队列中

        :param new_request:
        :param old_request:
        :return:
        """
        # 自动拼接 url（根据上一个请求的域名）
        if not new_request.url.startswith('http'):
            prefix = urlparse(old_request.url)

            if new_request.url.startswith('//'):
                new_request.url = new_request.url.lstrip('/')
                new_request.url = prefix.scheme + '://' + new_request.url
            else:
                new_request.url = new_request.url.lstrip('/')
                new_request.url = prefix.scheme + '://' + prefix.netloc + '/' + new_request.url

        new_request.callback = new_request.callback.__name__  # 转成字符串，不然无法序列化
        new_request.session = old_request.session  # 续上上一个的 session
        new_request.cookie_jar = old_request.cookie_jar  # 续上上一个的 cookie_jar
        self.queue.put(new_request)

    def run_requests(self, request: Request):
        """
        处理 yield 的 Request

        :param request:
        :return:
        """
        # 请求发起前的处理
        for middleware in self.__class__.REQUEST_MIDDLEWARE:
            middleware.request_in(self.spider, request)

        # 处理请求
        callback = None
        response = None
        failed_times = 0
        while failed_times < settings.REQUEST_RETRY_TIMES:
            try:
                callback, response = request.send()
                break
            except exception_drop.DropRequestException:
                raise
            except Exception as e:
                failed_times += 1

                # 出现错误可直接处理，并返回新的请求，同时阻断当前错误请求传播
                for middleware in self.__class__.REQUEST_MIDDLEWARE:
                    new_request = middleware.request_error(self.spider, request, e.__class__.__name__,
                                                           traceback.format_exc())
                    if new_request is None:
                        pass
                    elif isinstance(new_request, Request):
                        self.add_new_request(new_request=new_request, old_request=request)
                        raise exception_drop.DropRequestException()
                    else:
                        logger.warning("request_error 仅支持 Request 返回值！")

            # 失败达到最大次数直接丢弃
            if failed_times >= settings.REQUEST_RETRY_TIMES:
                for middleware in self.__class__.REQUEST_MIDDLEWARE:
                    middleware.request_failed(self.spider, request)
                return

        # 请求结束的回调（可在此时判断返回值是否合理，不合理可在此重新构造请求，同时阻断当前请求传播）
        for middleware in self.__class__.REQUEST_MIDDLEWARE:
            new_request = middleware.request_close(self.spider, request, response)
            if new_request is None:
                pass
            elif isinstance(new_request, Request):
                self.add_new_request(new_request=new_request, old_request=request)
                raise exception_drop.DropRequestException()
            else:
                logger.warning("request_close 仅支持 Request 返回值！")

        # 继续监控下次yield
        if callback is not None:
            if inspect.isgeneratorfunction(callback):  # 很好用的库，判断是否是 yield 函数
                for task in callback(request, response):
                    self.parse_task(task=task, request=request)
            else:
                callback(request, response)



"""
    返回带有解析器功能的 Response
"""

import re
import time
import requests
from typing import Any
from palp import settings
from parsel import Selector
from requests.cookies import RequestsCookieJar
from requests.models import Response as RequestResponse


# 响应解析器
class Response(RequestResponse):
    def __init__(self, resp: RequestResponse):
        """

        :param resp: requests 返回的响应
        """
        super().__init__()
        self._resp = resp
        self._resp.encoding = resp.apparent_encoding
        self._selector = Selector(self._resp.text)

        self.__dict__.update(self._resp.__dict__)  # 把 request Response 对象的数据转移过来

    def json(self, **kwargs) -> Any:
        return self._resp.json(**kwargs)

    def xpath(self, query: str, namespaces: str = None, **kwargs):
        """
        实现 xpath 选择器功能

        :param query: 匹配规则
        :param namespaces: 命名空间
        :param kwargs:
        :return:
        """
        return self._selector.xpath(query=query, namespaces=namespaces, **kwargs)

    def css(self, query: str):
        """
        实现 css 选择器功能

        :param query: 匹配规则
        :return:
        """
        return self._selector.css(query=query)

    def re(self, pattern: str, flags=0) -> list:
        """
        常用 flags：
            re.I：忽略字母大小写
            re.A：匹配相应的 ASCII 字符类别

        :param pattern: 匹配规则
        :param flags: 修饰符
        :return:
        """
        return re.findall(pattern, self.text, flags=flags)

    def re_first(self, pattern: str, flags=0) -> str:
        """
        常用 flags：
            re.I：忽略字母大小写
            re.A：匹配相应的 ASCII 字符类别

        :param pattern: 匹配规则
        :param flags: 修饰符
        :return:
        """
        result = self.re(pattern=pattern, flags=flags)
        if result:
            return result[0]

    def __setattr__(self, key, value):
        self.__dict__[key] = value

    def __getattr__(self, item):
        return self.__dict__.get(item)


# 发起请求，返回响应，含有 session 功能
class ResponseDownloader:

    def __init__(self, keep_session: bool, session: requests.Session, cookie_jar: RequestsCookieJar):
        """

        :param keep_session: 是否需要保持 session
        :param session: session
        :param cookie_jar: cookie_jar
        """
        self.keep_session = keep_session
        self.session = session
        self.cookie_jar = cookie_jar

    def response(self, **kwargs) -> Response:
        time.sleep(settings.REQUEST_DELAY)  # 请求等待延迟

        if kwargs.get('cookies'):
            self.cookie_jar.update(kwargs['cookies'])

        if self.keep_session:
            kwargs.update({'cookies': self.cookie_jar})

        resp = self.session.request(**kwargs)

        # 更新请求的 cookie、响应的 cookie
        self.cookie_jar.update(resp.cookies.get_dict())
        self.session.cookies.clear()

        return Response(resp)



from palp.middleware.middleware_spider_base import BaseSpiderMiddleware
from palp.middleware.middleware_request_base import BaseRequestMiddleware
from palp.middleware.middleware_request_filter import RequestBloomFilterMiddleware, RequestMemoryFilterMiddleware, \
    RequestRedisBloomFilterMiddleware, RequestRedisFilterMiddleware



"""
    pipeline
"""
from loguru import logger
from typing import Union, List
from palp.item.item_base import BaseItem
from palp.spider.spider_base import BaseSpider


class BasePipeline:
    def pipeline_in(self, spider: BaseSpider, item: BaseItem) -> None:
        """
        入库之前的操作

        :param spider:
        :param item:
        :return:
        """
        pass

    def pipeline_save(self, spider: BaseSpider, item: Union[BaseItem, List[BaseItem]]) -> None:
        """
        入库

        :param spider:
        :param item:
        :return:
        """
        pass

    def pipeline_error(self, spider: BaseSpider, item: Union[BaseItem, List[BaseItem]], exception_type: str,
                       exception: str) -> None:
        """
        入库出错时的操作

        :param spider:
        :param item: 启用 item buffer 时是 List[BaseItem]
        :param exception_type: 错误的类型
        :param exception: 错误的详细信息
        :return:
        """
        logger.error(exception)

    def pipeline_failed(self, spider: BaseSpider, item: Union[BaseItem, List[BaseItem]]) -> None:
        """
        超过最大重试次数时的操作

        :param spider:
        :param item: 启用 item buffer 时是 List[BaseItem]
        :return:
        """
        pass

    def pipeline_close(self, spider: BaseSpider) -> None:
        """
        spider 结束时的操作

        :param spider:
        :return:
        """
        pass